#!/usr/bin/env python3
# -*- coding:utf-8 -*-

from datetime import datetime
from typing import Dict, List, Optional

from pydantic import BaseModel


class Engine(BaseModel):
    """
    Retrosynthesis engine data model.
    """

    id: str
    name: str
    last_alive: datetime
    default: bool


class Library(BaseModel):
    """
    Retrosynthesis library data model.
    """

    id: str
    name: str
    version: str
    # available_from: datetime # NOTE: Masked from JSON output


class _Parameters(BaseModel):
    """
    Retrosynthesis job parameters data model.
    """

    retrosynthesis_engine: str
    building_block_libraries: List[str]
    number_of_routes: int
    processing_time: int
    reaction_limit: int
    building_block_limit: int


class _Step(BaseModel):
    """
    Retrosynthesis job route step data model.
    """

    order: int
    reaction_smiles: str


class Route(BaseModel):
    """
    Retrosynthesis job route data model.
    """

    summary: str
    building_blocks: List[Dict]
    steps: List[_Step]


class Job(BaseModel):
    """
    Retrosynthesis job data model.
    """

    id: str
    query: str
    status: str
    created: datetime
    updated: datetime
    routes: List[Route]
    parameters: _Parameters
    tags: Optional[List[str]]


class JobPage(BaseModel):
    """
    Retrosynthesis job paginated result data model.
    """

    results: List[Job]
    count: int
    limit: int


class NewJob(BaseModel):
    """
    Retrosynthesis job submission data model.
    """

    id: str
    smiles: str



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

from pendingai.context.cache import Cache
from pendingai.context.config import Config, Namespace
from pendingai.context.context import Context

__all__ = ["Cache", "Namespace", "Config", "Context"]



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

from typing import Callable, Dict, Optional

from pendingai.auth.flows import DeviceAuthorizationFlow, RefreshTokenFlow
from pendingai.auth.models import AccessToken
from rich.console import Console
from typer import Context, Exit, Typer

command: Typer = Typer(
    name="auth",
    no_args_is_help=True,
    add_completion=False,
    context_settings={"max_content_width": 90},
)

console: Console = Console(width=90)
print: Callable[..., None] = console.print


@command.callback()
def _callback(ctx: Context) -> None:
    """
    Authenticate with the Pending AI platform.
    """


@command.command("login")
def _login(ctx: Context) -> None:
    """
    Login to a Pending AI account. Device authorization will open a
    web browser window.

    For any account problems, contact support via <pendingai support>.
    """
    if ctx.obj.api_key is not None:
        print("[yellow]![/] Optional API key provided.")
        return

    auth_params: Dict[str, str] = ctx.obj.config.auth.model_dump()

    if ctx.obj.cache.access_token is not None:
        access_token: Optional[AccessToken] = ctx.obj.cache.access_token
        if access_token is not None and not access_token.is_expired():
            user_email: str = access_token.get_email()
            print("[yellow]! Authentication already complete.")
            print(f"[green]✓[/] Logged in as [b]{user_email}[/]")
            return
        else:
            refresh_flow: RefreshTokenFlow = RefreshTokenFlow(**auth_params)
            print("[yellow]! Authentication has expired, refreshing access.")
            if access_token is not None:
                access_token = refresh_flow.refresh(access_token.refresh_token)
            if access_token is None:
                print("[yellow]! Failed to refresh access, re-authenticating now.")
            else:
                print(f"[green]✓[/] Logged in as [b]{access_token.get_email()}[/]")
                ctx.obj.cache.access_token = access_token
                ctx.obj.cache.save(ctx.obj.environment)
                return

    device_auth: DeviceAuthorizationFlow = DeviceAuthorizationFlow(**auth_params)
    ctx.obj.cache.access_token = device_auth.authorize()
    ctx.obj.cache.save(ctx.obj.environment)
    print("[green]✓[/] Authentication complete.")
    print(f"[green]✓[/] Logged in as [b]{ctx.obj.cache.access_token.get_email()}[/]")


@command.command("logout")
def _logout(ctx: Context) -> None:
    """
    Logout of a Pending AI account. Removes locally cached
    authentication details.
    """
    if ctx.obj.cache.access_token:
        email: str = ctx.obj.cache.access_token.get_email()
        ctx.obj.cache.delete(ctx.obj.environment)
        print(f"[green]✓[/] Logged out [b]{email}[/]")
    else:
        print("[yellow]! Already logged out.")


@command.command("refresh", hidden=True)
def _refresh(ctx: Context) -> None:
    """
    TODO: Finish documentation.
    """


@command.command("status", hidden=True)
def _status(ctx: Context) -> None:
    """
    TODO: Finish documentation.
    """


@command.command("token")
def _token(ctx: Context) -> None:
    """
    Print the authentication token Pending AI uses for an account.
    """
    if ctx.obj.cache is None or ctx.obj.cache.access_token is None:
        print(
            "[yellow]! Use [b]<pendingai auth login>[/] to generate an authentication token."
        )
        raise Exit(1)
    elif ctx.obj.cache.access_token.access_token is None:
        print(
            "[yellow]! Invalid authentication token; use [b]<pendingai auth logout>[/]."
        )
        raise Exit(1)
    elif ctx.obj.cache.access_token.is_expired():
        print(
            "[yellow]! Authentication token has expired; use [b]<pendingai auth login>[/]."
        )
        raise Exit(1)
    else:
        user_email: str = ctx.obj.cache.access_token.get_email()
        print(f"[green]✓[/] Access token for [b]{user_email}[/]")
        print(f"\b{ctx.obj.cache.access_token.access_token}", soft_wrap=True)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import sys
from abc import ABC
from typing import Callable

from httpx import RequestError
from pydantic import ValidationError
from rich import print
from typer import BadParameter, Context, Exit


def capture_controller_errors(method: Callable) -> Callable:
    """
    Capture exceptions and errors that occur when interfacing with
    a command controller instance.

    Args:
        method (Callable): Controller method with error handling.

    Returns:
        Callable: Decorator method with controller error handling.
    """

    def controller_method(*args, **kwargs):
        try:
            return method(*args, **kwargs)
        except BadParameter as ex:
            raise ex
        except ValidationError:
            print(
                "[red]\u2717",
                "[red]Unexpected response data from the service, contact support",
                "[red]for more information.",
            )
        except RequestError:
            print(
                "[red]\u2717",
                "[red]Service request failed due to network or data transport",
                "[red]errors, try again or see the relevant documentation.",
            )
        except Exit as e:
            raise e
        except Exception:
            print(
                "[red]\u2717",
                "[red]Unexpected error encountered by the service, contact support",
                "[red]for more information.",
            )
        sys.exit(1)

    return controller_method


class Controller(ABC):
    """
    Abstract superclass for interfacing with command-specific logic.

    Attributes:
        ctx (typer.Context): Runtime application context instance.
    """

    def __init__(self, ctx: Context):
        self.ctx: Context = ctx



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

from __future__ import annotations

import json
import re
import unicodedata
from datetime import datetime
from typing import Annotated, Any, Dict, Generator, List

from pendingai.commands import append_client_context
from pendingai.commands.retrosynthesis.controller import Retrosynthesis
from pendingai.commands.retrosynthesis.models import (
    Engine,
    Job,
    JobPage,
    Library,
    NewJob,
)
from pendingai.commands.retrosynthesis.types import Options
from rich import print, print_json
from rich.console import Console
from rich.prompt import Confirm
from rich.status import Status
from rich.table import Table, box
from typer import Argument, BadParameter, Context, Exit, Option, Typer

app: Typer = Typer(context_settings={"max_content_width": 90})


@app.callback()
def _callback(ctx: Context) -> None:
    """
    High-throughput deep-learning retrosynthesis service.

    Retrosynthesis predictions are generated by submitting jobs to an
    engine, monitoring their status, and retrieving the completed results.
    For more information refer to the documentation with <pendingai docs>.
    """


@app.command("engines")
def _engines(
    ctx: Context,
    render_json: Options.RenderJson = False,
) -> None:
    """
    List available retrosynthesis engines. Their unique IDs are required
    for submitting jobs. If no engines are available, contact support
    with <pendingai support>.
    """
    append_client_context(ctx)
    ctx.obj.controller = Retrosynthesis(ctx)

    engines: List[Engine] = ctx.obj.controller.get_engines()
    engines.sort(key=lambda x: x.last_alive, reverse=True)
    engines.sort(key=lambda x: x.default, reverse=True)

    if len(engines) == 0:
        print("[yellow]! No retrosynthesis engine available.")
        raise Exit(1)

    if render_json:
        print_json(data=[engine.model_dump(mode="json") for engine in engines])

    else:
        table: Table = Table(box=box.SQUARE)
        table.add_column("ID", style="bold")
        table.add_column("Name")
        table.add_column("Last Alive", style="dim", justify="right")
        for engine in engines:
            ts: float = (datetime.utcnow() - engine.last_alive).seconds
            table.add_row(
                engine.id + (" (default)" if engine.default else ""),
                engine.name,
                (f"{ts:3d}" if ts < 60 else ">60") + " seconds ago",
            )
        print(table)


@app.command("libraries")
def _libraries(
    ctx: Context,
    render_json: Options.RenderJson = False,
) -> None:
    """
    List available building block libraries. Their unique IDs are required
    for submitting jobs. If no libraries are available, contact support
    with <pendingai support>.
    """
    append_client_context(ctx)
    ctx.obj.controller = Retrosynthesis(ctx)

    libraries: List[Library] = ctx.obj.controller.get_libraries()
    libraries.sort(key=lambda x: x.name)

    if len(libraries) == 0:
        print("[yellow]! No building block libraries available.")
        raise Exit(1)

    if render_json:
        print_json(data=[library.model_dump(mode="json") for library in libraries])

    else:
        table: Table = Table(box=box.SQUARE)
        table.add_column("ID", style="bold")
        table.add_column("Name")
        table.add_column("Version", style="dim")
        for library in libraries:
            table.add_row(library.id, library.name, library.version)
        print(table)


def _sanitize_filename(pth: str) -> str:
    """
    TODO: Finish documentation.
    """
    if "." in pth:
        pth = ".".join(pth.split(".")[:-1])
    pth = unicodedata.normalize("NFKD", pth).encode("ascii", "ignore").decode("ascii")
    pth = re.sub(r"[^\.\w\s-]", "", pth.lower())
    return re.sub(r"[-\.\s]+", "_", pth).strip("-_")


@app.command("submit", no_args_is_help=True)
def _submit(
    ctx: Context,
    output_file: Options.OutputIdFile,
    targets_list: Options.InputTargetsList = [],
    targets_file: Options.InputTargetsFile = None,
    eng_id: Options.Parameters.EngineId = None,
    lib_ids: Options.Parameters.LibraryIds = None,
    number_of_routes: Options.Parameters.NumRoutes = 20,
    processing_time: Options.Parameters.ProcessingTime = 300,
    reaction_limit: Options.Parameters.ReactionLimit = 8,
    building_block_limit: Options.Parameters.BlockLimit = 8,
    job_tags: Options.Parameters.JobTags = [],
) -> None:
    """
    Submit jobs to retrosynthesis engine. Writes IDs for submitted jobs
    to output file. Jobs are automatically tagged with submission
    time and output filename.

    WARNING: Jobs including their results will be automatically deleted
    30 days after completion.

    \b
    Tags and building block libraries are repeatable:
      <pendingai retro submit --output-file job1.ids --tag tag1 --tag tag2>
      <pendingai retro submit --output-file job1.ids --lib-id lib1 --lib-id lib2>
    """
    append_client_context(ctx)
    ctx.obj.controller = Retrosynthesis(ctx)

    console: Console = Console(quiet=False, highlighter=None)
    targets: List[str] = targets_list + (targets_file if targets_file else [])

    # Validate input values; targets must be provided and not exceed an
    # upper-limit to not overload the system.
    if len(targets) == 0:
        raise BadParameter("At least 1 input structure is required for retrosynthesis.")
    if len(targets) > 10_000:
        raise BadParameter("Exceeded request limit of 10,000.")
    console.print(f"[green]\u2713[/] Validated {len(targets)} retrosynthesis target(s).")

    job_tags.append(datetime.now().strftime("%Y%m%d-%H%M%S"))
    console.print(f'[green]\u2713[/] Added additional batch tag "{job_tags[-1]}"')
    job_tags.append(_sanitize_filename(output_file.name)[:16])
    console.print(f'[green]\u2713[/] Added additional batch tag "{job_tags[-1]}"')

    # Results are streamed back to the user in order to handle batched
    # requests and partial results. That way any incomplete requests are
    # still returned and end on a failed batch submission.
    results: Generator[List[NewJob], Any, None] = ctx.obj.controller.submit_molecules(
        targets=targets,
        engine=eng_id,
        libraries=lib_ids,
        number_of_routes=number_of_routes,
        processing_time=processing_time,
        reaction_limit=reaction_limit,
        building_block_limit=building_block_limit,
        tags=job_tags,
        console=console,
    )

    with open(output_file, "w") as fp:
        for batch in results:
            fp.writelines([result.id + "\n" for result in batch])


@app.command("delete", no_args_is_help=True)
def _delete(
    ctx: Context,
    ids_list: Options.InputIdList = [],
    ids_file: Options.InputIdFile = None,
    output_file: Options.OutputFile = None,
    render_json: Options.RenderJson = False,
) -> None:
    """
    Delete one or more retrosynthesis jobs by ID. Deletion is irreversible,
    completed in order, and will skip any missing or failed IDs.

    WARNING: Jobs including their results will be automatically deleted 30 days
    after completion.
    """
    append_client_context(ctx)
    ctx.obj.controller = Retrosynthesis(ctx)

    if output_file is not None:
        output_file = open(output_file, "w")  # type: ignore

    console: Console = Console(quiet=render_json, highlighter=None)
    ids: List[str] = ids_list + (ids_file if ids_file else [])

    # Validate input values; targets must be provided and not exceed an
    # upper-limit to not overload the system.
    if len(ids) == 0:
        raise BadParameter("At least 1 job ID is required for deletion.")
    if len(ids) > 1000:
        raise BadParameter("Exceeded request limit of 1000.")

    # If the number of targets exceeds a limit of 25, there are too many
    # results to display in the terminal, this is then saved to a local
    # temporary filepath.
    if len(ids) > 25 and output_file is None:
        console.quiet = False
        path: str = f"delete_{datetime.now().strftime('%Y-%m-%dT%H:%M:%S')}.json"
        print(f"[yellow not b]! Over 25 IDs requested, saving results to file: {path}")
        output_file = open(path, "w")  # type: ignore

    results: Dict = ctx.obj.controller.delete_jobs(ids=ids, console=console)

    if output_file:
        json.dump(results, output_file, indent=2)  # type: ignore

    elif render_json:
        print_json(data=results)

    else:
        table: Table = Table("ID", "Status", box=box.SQUARE)
        [table.add_row(id, "[green]Success") for id in results["success"]]
        [table.add_row(id, "[red]Failed") for id in results["failure"]]
        [table.add_row(id, "[dim]Missing") for id in results["missing"]]
        print(table)


@app.command("status", no_args_is_help=True)
def _status(
    ctx: Context,
    ids_list: Options.InputIdList = [],
    ids_file: Options.InputIdFile = None,
    output_file: Options.OutputFile = None,
    render_json: Options.RenderJson = False,
) -> None:
    """
    Get the status of one or more retrosynthesis jobs by ID.
    """
    append_client_context(ctx)
    ctx.obj.controller = Retrosynthesis(ctx)

    if output_file is not None:
        output_file = open(output_file, "w")  # type: ignore

    console: Console = Console(quiet=render_json, highlighter=None)
    ids: List[str] = ids_list + (ids_file if ids_file else [])

    # Validate input values; targets must be provided and not exceed an
    # upper-limit to not overload the system.
    if len(ids) == 0:
        raise BadParameter("At least 1 job ID is required for checking a status.")
    if len(ids) > 1000:
        raise BadParameter("Exceeded request limit of 1000.")

    # If the number of targets exceeds a limit of 25, there are too many
    # results to display in the terminal, this is then saved to a local
    # temporary filepath.
    if len(ids) > 25 and output_file is None:
        console.quiet = False
        path: str = f"status_{datetime.now().strftime('%Y-%m-%dT%H:%M:%S')}.json"
        print(f"[yellow not b]! Over 25 IDs requested, saving results to file: {path}")
        output_file = open(path, "w")  # type: ignore

    results: List[Dict] = ctx.obj.controller.get_job_status(ids=ids, console=console)

    if output_file:
        json.dump(results, output_file, indent=2)  # type: ignore

    elif render_json:
        print_json(data=results)

    else:
        table: Table = Table("ID", "Status", box=box.SQUARE)
        for result in results:
            if result["status"] == "completed":
                result["status"] = "[green]Completed"
            elif result["status"] == "submitted":
                result["status"] = "Submitted"
            elif result["status"] == "processing":
                result["status"] = "[yellow]Processing"
            elif result["status"] == "missing":
                result["status"] = "[dim]Missing"
            else:
                result["status"] = "[red]Failed"
            table.add_row(result["id"], result["status"])
        print(table)


@app.command("view", no_args_is_help=True)
def _view(
    ctx: Context,
    ids_list: Options.InputIdList = [],
    ids_file: Options.InputIdFile = None,
    output_file: Options.OutputFile = None,  # type: ignore
    render_json: Options.RenderJson = False,
) -> None:
    """
    Retrieve retrosynthesis job results by ID. IDs are processed
    sequentially. Retrosynthetic routes are only returned in JSON output.
    """
    append_client_context(ctx)
    ctx.obj.controller = Retrosynthesis(ctx)

    if output_file is not None:
        output_file = open(output_file, "w")  # type: ignore

    console: Console = Console(quiet=render_json, highlighter=None)
    ids: List[str] = ids_list + (ids_file if ids_file else [])

    # Validate input values; targets must be provided and not exceed an
    # upper-limit to not overload the system.
    if len(ids) == 0:
        raise BadParameter("At least 1 job ID is required for retrieving results.")
    if len(ids) > 10_000:
        raise BadParameter("Exceeded request limit of 10,000.")

    # If the number of targets exceeds a limit of 25, there are too many
    # results to display in the terminal, this is then saved to a local
    # temporary filepath.
    if len(ids) > 25 and output_file is None:
        console.quiet = False
        path: str = f"results_{datetime.now().strftime('%Y-%m-%dT%H:%M:%S')}.json"
        print(f"[yellow not b]! Over 10 IDs requested, saving results to file: {path}")
        output_file = open(path, "w")  # type: ignore

    results: Generator[Job | Dict, Any, None] = ctx.obj.controller.get_jobs_by_id(
        ids=ids,
        console=console,
    )

    # Special file writing is needed to comply with JSON formatting and
    # can handle different response formats depending on if the job ID
    # has existing results or does not.
    if output_file:
        output_file.write("[\n\t")  # type: ignore
        for i, result in enumerate(results):
            if i > 0:
                output_file.write(",\n\t")  # type: ignore
            try:
                output_file.write(result.model_dump_json())  # type: ignore
            except Exception:
                output_file.write(json.dumps(result))  # type: ignore
        output_file.write("\n]")  # type: ignore

    elif render_json:
        print_json(
            data=[
                x.model_dump(mode="json") if not isinstance(x, dict) else x
                for x in results
            ]
        )

    else:
        table: Table = Table(box=box.SQUARE)
        table.add_column("ID", width=24)
        table.add_column("Structure", width=50, style="dim")
        table.add_column("Status")
        table.add_column("Routes", justify="right")
        results_: list = [i for i in results]
        has_tabular_view: bool = False
        for result in results_:
            if isinstance(result, dict):
                print(f"[yellow not b]! Job is not completed or missing: {result['id']}")
            else:
                has_tabular_view = True
                table.add_row(
                    result.id,
                    result.query,
                    result.status.title(),
                    str(len(result.routes)) if len(result.routes) else "0",
                )
        print("[yellow]! Use --json to retrieve detailed retrosynthetic routes.")
        if has_tabular_view:
            print(table)


@app.command("list")
def _list(
    ctx: Context,
    page: Options.Pagination.PageNumber = 1,
    page_size: Options.Pagination.PageSize = 10,
    status: Options.Pagination.Status = None,
    tags: Options.Parameters.Tags = [],
    render_json: Options.RenderJson = False,
) -> None:
    """
    Retrieve a page of retrosynthesis jobs.
    Displays select search parameters and a binary synthesizability flag.
    Use <pendingai retro view> to retrieve retrosynthetic routes of
    completed jobs.

    \b
    Tags are repeatable:
      <pendingai retro list --tag tag1 --tag tag2>
    """
    append_client_context(ctx)
    ctx.obj.controller = Retrosynthesis(ctx)

    console: Console = Console(quiet=render_json, highlighter=None)
    page_data: JobPage = ctx.obj.controller.get_paginated_jobs(
        page=page,
        page_size=page_size,
        status=status.value if status else None,
        tags=tags,
    )

    if page_data.limit > 0 and page > page_data.limit:
        console.print(f"[yellow]! Page limit of {page_data.limit} exceeded.")
        raise Exit(1)
    elif len(page_data.results) == 0:
        console.print("[yellow]! No results found with the given filter(s).")
        raise Exit(1)

    if render_json:
        print_json(data=[x.model_dump(mode="json") for x in page_data.results])

    else:
        table: Table = Table(box=box.SQUARE, caption=f"Page {page} of {page_data.limit}")
        table.add_column("ID", style="bold green")
        table.add_column("Structure", width=50)
        table.add_column("Submitted At", style="dim")
        table.add_column("Status", style="cyan")
        table.add_column("Engine", width=15, style="dim")
        table.add_column("Libraries", width=15, style="dim", no_wrap=True)
        table.add_column("Tags", width=20)
        table.add_column("Synthesizable", style="bold cyan")
        for result in page_data.results:
            table.add_row(
                result.id,
                result.query,
                result.created.strftime("%Y-%m-%d %H:%M:%S"),
                result.status.title(),
                result.parameters.retrosynthesis_engine,
                ", ".join(result.parameters.building_block_libraries),
                ", ".join(result.tags) if result.tags else "",
                str(len(result.routes) > 0) if result.status == "completed" else "",
            )
        print(table)


def _tag_list_callback(ctx: Context, value: bool) -> None:
    if value:
        append_client_context(ctx)
        ctx.obj.controller = Retrosynthesis(ctx)
        with Status("Retrieving all job tags"):
            tag_ids: list[str] = ctx.obj.controller.get_job_tags()

        if len(tag_ids) == 0:
            print("[yellow]! No job tags found.")
        else:
            print(f"Retrieved {len(tag_ids)} job tags:")
            for tag_id in tag_ids:
                print(f"- '{tag_id}'")
        raise Exit(0)


@app.command("tag")
def _tag(
    ctx: Context,
    tag_id: Annotated[
        str,
        Argument(
            metavar="TAG",
            help="The operations is carried out on this single tag.",
        ),
    ],
    list_tags: Annotated[
        bool,
        Option(
            "--list",
            help="List all job tags.",
            callback=_tag_list_callback,
            is_eager=True,
        ),
    ] = False,
    bulk_delete_tag_flag: Annotated[
        bool,
        Option(
            "--delete",
            help="Delete jobs with matching tag.",
            is_flag=True,
        ),
    ] = False,
    retrieve_tag_count_flag: Annotated[
        bool,
        Option(
            "--count",
            help="Count jobs with matching tag.",
            is_flag=True,
        ),
    ] = False,
) -> None:
    """
    Tag-based operations.
    """
    append_client_context(ctx)
    ctx.obj.controller = Retrosynthesis(ctx)

    console = Console(highlighter=None)

    # Display a wait status while performing the count operation
    with Status(f"Finding jobs with tag: '{tag_id}'"):
        count: int = ctx.obj.controller.get_job_tag_count(tag_id)

    if bulk_delete_tag_flag:
        # Confirm the user wants to delete all jobs with the tag
        print(f"[yellow not b]! Warning: Deleting {count} jobs tagged with '{tag_id}'.")
        print("[yellow]- This action cannot be undone.")
        if Confirm.ask("[yellow]! Are you sure?"):
            # Run the delete operation and output the number deleted
            with Status(f"Deleting jobs with tag: '{tag_id}'"):
                deleted: int = ctx.obj.controller.delete_job_tag(tag_id)
            console.print(f"[green]✓[/] Deleted {deleted} jobs with tag: '{tag_id}'")
        raise Exit(0)

    if retrieve_tag_count_flag:
        # Output the retrieved count
        console.print(f"[green]✓[/] Found {count} job(s) with tag: '{tag_id}'")
        raise Exit(0)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

from typing import Optional

from pydantic import BaseModel, ConfigDict, computed_field, model_validator

from pendingai import Environment
from pendingai.client import Client
from pendingai.commands.controller import Controller
from pendingai.context.cache import Cache
from pendingai.context.config import Config


class Context(BaseModel):
    """
    Runtime context data model for typer object state.
    """

    app_name: str
    environment: Environment
    subcommand: str
    api_key: Optional[str] = None
    client: Optional[Client] = None
    cache: Cache = None  # type: ignore
    config: Config = None  # type: ignore
    controller: Controller = None  # type: ignore

    model_config: ConfigDict = ConfigDict(arbitrary_types_allowed=True)

    @model_validator(mode="after")
    def init_computed_fields(self) -> "Context":
        """
        Initialise null-defined context fields on startup.
        """
        self.cache = Cache.load(self.environment)
        self.config = Config.load(self.environment)
        return self

    @computed_field  # type: ignore
    @property
    def session_token(self) -> Optional[str]:
        """
        Runtime session token determined by the api key or access token.
        """
        if self.api_key:
            return self.api_key
        if self.cache.access_token:
            return self.cache.access_token.access_token
        return None



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

from pathlib import Path
from typing import Any, Dict

from pydantic import BaseModel, ValidationError
from yaml import safe_load

from pendingai import Environment


class _Command(BaseModel):
    """
    Namespace command data model.
    """

    subcommand: str
    api_domain: str


class _CommandMapping(BaseModel):
    """
    Namespace command mapping data model.
    """

    retrosynthesis: _Command


class Namespace(BaseModel):
    """
    Namespace data model.
    """

    commands: _CommandMapping
    support_redirect_url: str
    docs_redirect_url: str
    labs_redirect_url: str

    @classmethod
    def load(cls) -> "Namespace":
        """
        Custom model loader from static file.

        Raises:
            RuntimeError: Invalid configuration file.

        Returns:
            Namespace: Loaded class instance.
        """
        try:
            path: Path = Path(__file__).parent / "globals.yml"
            contents: Dict[str, Any] = safe_load(path.open())
            return cls.model_validate(contents)

        except (ValidationError, FileExistsError, FileNotFoundError) as ex:
            raise RuntimeError("Invalid application state.") from ex


class _ApiEnvironment(BaseModel):
    """
    Config api environment model.
    """

    domain: str


class _AuthEnvironment(BaseModel):
    """
    Config authentication environment data model.
    """

    domain: str
    client_id: str
    audience: str


class Config(BaseModel):
    """
    Config data model.
    """

    api: _ApiEnvironment
    auth: _AuthEnvironment
    environment: Environment

    @classmethod
    def load(cls, environment: Environment = Environment.PRODUCTION) -> "Config":
        """
        Custom model loader from static file for a specific environment.

        Args:
            environment (Environment, optional): Instance environment.

        Raises:
            RuntimeError: Invalid configuration file.

        Returns:
            Namespace: Loaded class instance.
        """
        try:
            path: Path = Path(__file__).parent / f"config.{environment.value}.yml"
            contents: Dict[str, Any] = safe_load(path.open())
            contents["environment"] = environment
            return cls.model_validate(contents)

        except (ValidationError, FileExistsError, FileNotFoundError) as ex:
            raise RuntimeError("Invalid application state.") from ex



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

from rich import print
from typer import Context, Exit

from pendingai.client import Client
from pendingai.context import Namespace

namespace: Namespace = Namespace.load()


def append_client_context(ctx: Context) -> None:
    """
    Update the app runtime context with an initialised client. Required
    context fields are validated and the command `/alive` endpoint is
    hit to confirm the service is available.

    Args:
        ctx (Context): App runtime context.
        command (str): Command with matching client domain.
    """

    # Check profile session token exists
    if ctx.obj.session_token is None:
        print("[yellow]! No authenticated session found, try logging in.")
        raise Exit(0)

    # Check session token is not expired
    if ctx.obj.api_key is None and ctx.obj.cache.access_token.is_expired():
        print("[yellow]! Authenticated session has expired, try logging in.")
        raise Exit(0)

    # Initialise only the retrosynthesis command
    ctx.obj.client = Client(
        domain=ctx.obj.config.api.domain,
        subdomain=namespace.commands.retrosynthesis.api_domain,
        token=ctx.obj.session_token,
    )

    # Check client is reachable
    if ctx.obj.client.get("/alive").status_code != 200:
        print(
            f"[red]\u2717 Pending AI's <{ctx.command.name}> Service is currently",
            "[red]unavailable, try again shortly or contact support with",
            "[red][b]<pendingai support>[/].",
        )
        raise Exit(1)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

from datetime import datetime, timezone
from typing import Dict, Optional

import jwt
from pydantic import BaseModel

_namespace: str = "https://pending.ai/claims"


def get_jwt_token_claims(token: str) -> Dict:
    """
    Extract a jwt token claims segment without verifying the token
    signature. No external http request is made to the known jwks.

    Args:
        token (str): Jwt token to decode using `jwt`.

    Raises:
        ValueError: Exception encountered decoding the jwt token.

    Returns:
        Dict: Jwt token claims.
    """
    try:
        return jwt.decode(token, options={"verify_signature": False})
    except Exception as ex:
        raise ValueError("Cannot decode invalid token.") from ex


def access_token_is_expired(token: str) -> bool:
    """
    Check if an access token has reached its expiry timestamp.

    Args:
        token (str): Access token being checked for expiration.

    Raises:
        ValueError: Exception encountered decoding the jwt token.

    Returns:
        bool: Access token is expired.
    """
    try:
        tz: timezone = timezone.utc
        expires_at: int = get_jwt_token_claims(token)["exp"]
        return datetime.fromtimestamp(expires_at, tz=tz) < datetime.now(tz=tz)
    except Exception as ex:
        raise ValueError("Cannot decode invalid token.") from ex


class AccessToken(BaseModel):
    """
    Access token abstraction as received from Auth0 using the device
    authorization code or refresh token flows with `/oauth/token`.
    Helper methods are defined for checking if the token is expired or
    to extract additional jwt claims.
    """

    access_token: str
    refresh_token: str
    id_token: str
    token_type: str
    expires_in: int
    scope: str

    def get_email(self) -> str:
        """
        Extract the custom user email claim from the access token.

        Returns:
            str: User email claim.
        """
        return get_jwt_token_claims(self.access_token)[f"{_namespace}/email"]

    def get_account_type(self) -> str:
        """
        Extract the custom account type claim from the access token.

        Returns:
            str: Account type claim.
        """
        return get_jwt_token_claims(self.access_token)[f"{_namespace}/account_type"]

    def get_org_id(self) -> Optional[str]:
        """
        Extract the custom org id claim from the access token.

        Returns:
            str: Org id claim.
        """
        return get_jwt_token_claims(self.access_token).get(f"{_namespace}/org_id")

    def get_org_name(self) -> Optional[str]:
        """
        Extract the custom org name claim from the access token.

        Returns:
            str: Org name claim.
        """
        return get_jwt_token_claims(self.access_token).get(f"{_namespace}/org_name")

    def is_expired(self) -> bool:
        """
        Check if the access_token has expired.

        Returns:
            bool: _description_
        """
        return access_token_is_expired(self.access_token)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import re
from enum import Enum, unique
from io import TextIOWrapper
from pathlib import Path
from typing import Any, List, Optional

from pendingai.commands import append_client_context
from pendingai.commands.retrosynthesis.controller import Retrosynthesis
from rich.prompt import Confirm
from typer import BadParameter, Context, Exit, FileText, Option
from typing_extensions import Annotated

_JOB_TAG_LIMIT: int = 5
_JOB_TAG_REGEX: str = r"^[\w-]{,16}$"
_TARGET_LIMIT: int = 10_000
_TARGET_REGEX: str = r"^[A-Za-z0-9\(\)+-=#%+@\/\\\[\]]+$"
_IDS_LIMIT: int = 10_000
_IDS_REGEX: str = r"^[a-f\d]{24}$"


@unique
class Status(str, Enum):
    """
    Job status enum value choices.
    """

    SUBMITTED = "submitted"
    PROCESSING = "processing"
    COMPLETED = "completed"
    FAILED = "failed"


def _input_targets_list_callback(value: List[str]) -> List[str]:
    """
    Input targets list validation callback. Check the length of the list
    and check the string pattern for each individual target.

    Args:
        value (List[str]): Input target list.

    Raises:
        BadParameter: Targets list exceeds `_TARGET_LIMIT`.
        BadParameter: Targets do not match `_TARGET_REGEX`.

    Returns:
        List[str]: Input target list.
    """
    if len(value) > _TARGET_LIMIT:
        raise BadParameter(f"Exceeded job limit of {_TARGET_LIMIT}.")
    targets: list[str] = []
    for target_value in value:
        if not re.match(_TARGET_REGEX, target_value):
            raise BadParameter("Query structures must be a valid SMILES.")
        if len(target_value) > 512:
            print(
                f"Query structures must be less than 512 chars, skipping: {target_value}"
            )
            continue
        targets.append(target_value)
    return targets


def _input_targets_file_callback(value: Optional[TextIOWrapper]) -> List[str]:
    """
    Input targets file validation callback. Check the length of the file
    and check the string pattern for each individual target.

    Args:
        value (Optional[TextIOWrapper]): Input target file.

    Raises:
        BadParameter: Targets file exceeds `_TARGET_LIMIT`.
        BadParameter: Targets do not match `_TARGET_REGEX`.

    Returns:
        List[str]: Input target file converted to a list.
    """
    line_values: List[str] = []
    if value is not None:
        for line_no, line_value in enumerate(value):
            if line_no > _TARGET_LIMIT:
                raise BadParameter(f"Exceeded job limit of {_TARGET_LIMIT}.")
            line_values.append(line_value.strip())
            if not re.match(_TARGET_REGEX, line_values[-1]):
                raise BadParameter("Query structures must be a valid SMILES.")
            if len(line_values[-1]) > 512:
                print(
                    f"Query structures must be less than 512 chars, skipping: {line_values[-1]}"
                )
                line_values.pop()
    return line_values


def _input_id_list_callback(ctx: Context, value: List[str]) -> List[str]:
    """
    Input IDs list validation callback. Check the length of the list
    and check the string pattern for each individual IDs.

    Args:
        value (List[str]): Input IDs list.

    Raises:
        BadParameter: IDs list exceeds `_IDS_LIMIT`.
        BadParameter: IDs do not match `_IDS_REGEX`.

    Returns:
        List[str]: Input IDs list.
    """
    if len(value) > _IDS_LIMIT:
        raise BadParameter(f"Exceeded request limit of {_IDS_LIMIT}.")
    for id_value in value:
        if not re.match(_IDS_REGEX, id_value):
            raise BadParameter("Requested IDs have invalid format.")
    return value


def _input_id_file_callback(value: Optional[TextIOWrapper]) -> List[str]:
    """
    Input IDs file validation callback. Check the length of the file
    and check the string pattern for each individual ID.

    Args:
        value (Optional[TextIOWrapper]): Input IDs file.

    Raises:
        BadParameter: IDs file exceeds `_IDS_LIMIT`.
        BadParameter: IDs do not match `_IDS_REGEX`.

    Returns:
        List[str]: Input IDs file converted to a list.
    """
    line_values: List[str] = []
    if value is not None:
        for line_no, line_value in enumerate(value):
            if line_no > _IDS_LIMIT:
                raise BadParameter(f"Exceeded request limit of {_IDS_LIMIT}.")
            line_values.append(line_value.strip())
            if not re.match(_IDS_REGEX, line_values[-1]):
                raise BadParameter("Requested IDs have invalid format.")
    return line_values


def _library_callback(ctx: Context, libraries: Any) -> List[str]:
    """
    Check all building block library IDs are valid.

    Args:
        ctx (Context): Runtime app context.
        libraries (Any, optional): Building block library IDs.

    Raises:
        BadParameter: Library ID has invalid format.

    Returns:
        List[str]: Given library ID parameter values.
    """
    if libraries is not None:
        append_client_context(ctx)
        ctx.obj.controller = Retrosynthesis(ctx)
        ids: List[str] = [x.id for x in ctx.obj.controller.get_libraries()]
        if isinstance(libraries, str) and libraries in ids:
            return [libraries]
        elif isinstance(libraries, str):
            raise BadParameter(f"Building block library is not available: {libraries}")
        elif isinstance(libraries, list):
            for lib in libraries:
                if not isinstance(lib, str) or lib not in ids:
                    raise BadParameter(f"Building block library is not available: {lib}")
            return libraries
        raise BadParameter("Invalid parameter type.")
    return []


def _engine_callback(ctx: Context, engine: Any) -> Optional[str]:
    """
    Check a retrosynthesis engine ID is valid.

    Args:
        ctx (Context): Runtime app context.
        engine (Any, optional): Retrosynthesis engine ID.

    Raises:
        typer.BadParameter: Engine ID has invalid format.

    Returns:
        Optional[str]: Given engine ID parameter value.
    """
    if engine is not None and isinstance(engine, str):
        append_client_context(ctx)
        ctx.obj.controller = Retrosynthesis(ctx)
        ids: List[str] = [x.id for x in ctx.obj.controller.get_engines()]
        if engine not in ids:
            raise BadParameter(f"Retrosynthesis engine is not available: {engine}")
        return engine
    return None


def _eng_id_pagination_callback(ctx: Context, eng_id: Optional[str]) -> Optional[str]:
    """
    Check a retrosynthesis engine ID is valid.

    Args:
        ctx (Context): Runtime app context.
        eng_id (str, optional): Retrosynthesis engine ID.

    Raises:
        BadParameter: Engine ID is invalid.

    Returns:
        Optional[str]: Retrosynthesis engine ID.
    """
    if eng_id:
        append_client_context(ctx)
        ctx.obj.controller = Retrosynthesis(ctx)
        ids: List[str] = [x.id for x in ctx.obj.controller.get_engines()]
        if eng_id not in ids:
            raise BadParameter("Retrosynthesis engine is not available.")
    return eng_id


def _lib_id_pagination_callback(ctx: Context, lib_id: Optional[str]) -> Optional[str]:
    """
    Check a building block library ID is valid.

    Args:
        ctx (Context): Runtime app context.
        lib_id (str, optional): Building block library ID.

    Raises:
        BadParameter: Library ID is invalid.

    Returns:
        Optional[str]: Building block library ID.
    """
    if lib_id:
        append_client_context(ctx)
        ctx.obj.controller = Retrosynthesis(ctx)
        ids: List[str] = [x.id for x in ctx.obj.controller.get_libraries()]
        if lib_id not in ids:
            raise BadParameter("Building block library is not available.")
    return lib_id


def _tags_callback(value: List[str]) -> List[str]:
    """
    Job tags list validation callback. Check the length of the tags
    and check the string pattern for each individual tag.

    Args:
        value (List[str]): Input job tag list.

    Raises:
        BadParameter: Tags list exceeds `_JOB_TAG_LIMIT`.
        BadParameter: Tags do not match `_JOB_TAG_REGEX`.

    Returns:
        List[str]: Input job tag list.
    """
    if len(value) > _JOB_TAG_LIMIT:
        raise BadParameter(f"Maximum number of tags per job is {_JOB_TAG_LIMIT}.")
    for tag_value in value:
        if not re.match(_JOB_TAG_REGEX, tag_value):
            raise BadParameter("Tags must be alpha-numeric with underscores or dashes.")
    return value


def _job_tags_callback(value: List[str]) -> List[str]:
    """
    Job tags list validation callback. Check the length of the tags
    and check the string pattern for each individual tag.

    Args:
        value (List[str]): Input job tag list.

    Raises:
        BadParameter: Tags list exceeds `_JOB_TAG_LIMIT`.
        BadParameter: Tags do not match `_JOB_TAG_REGEX`.

    Returns:
        List[str]: Input job tag list.
    """
    if len(value) > 3:
        raise BadParameter(f"Maximum number of tags per job is {3}.")
    for tag_value in value:
        if not re.match(_JOB_TAG_REGEX, tag_value):
            raise BadParameter("Tags must be alpha-numeric with underscores or dashes.")
    return value


def _output_ids_filepath_callback(value: Path) -> Path:
    prompt: str = f"[yellow]? Are you sure you want to overwrite the file '{value}'"
    if value.exists() and not Confirm.ask(prompt):
        raise Exit(0)
    return value


def _output_results_filepath_callback(value: Path) -> Path:
    prompt: str = f"[yellow]? Are you sure you want to overwrite the file '{value}'"
    if value is not None and value.exists() and not Confirm.ask(prompt):
        raise Exit(0)
    return value


class Options:
    """
    Namespace for shared command options.
    """

    InputTargetsList = Annotated[
        List[str],
        Option(
            "--smiles",
            "-s",
            help="SMILES of the query structure. SMARTS are unsupported.",
            callback=_input_targets_list_callback,
        ),
    ]
    InputTargetsFile = Annotated[
        FileText,
        Option(
            "--input-file",
            "-i",
            help="Line-delimited file with SMILES of query structures.",
            callback=_input_targets_file_callback,
            exists=True,
        ),
    ]
    InputIdList = Annotated[
        List[str],
        Option(
            "--id",
            help="Retrosynthesis job IDs. An ID becomes available upon submitting a job. IDs can also be retrieved when listing jobs with <pendingai retro list>.",
            callback=_input_id_list_callback,
        ),
    ]
    InputIdFile = Annotated[
        Optional[FileText],
        Option(
            "--input-file",
            "-i",
            help="Line-delimited file with retrosynthesis job IDs.",
            callback=_input_id_file_callback,
            exists=True,
        ),
    ]
    OutputIdFile = Annotated[
        Path,
        Option(
            "--output-file",
            "-o",
            help="Output filepath for writing line-delimited job IDs.",
            callback=_output_ids_filepath_callback,
            exists=False,
            writable=True,
        ),
    ]
    OutputFile = Annotated[
        Optional[Path],
        Option(
            "--output-file",
            "-o",
            help="Output filepath for writing command results.",
            callback=_output_results_filepath_callback,
            exists=False,
            writable=True,
        ),
    ]
    RenderJson = Annotated[
        bool,
        Option(
            "--json",
            help="Render output as JSON.",
            is_flag=True,
        ),
    ]
    RouteNumber = Annotated[
        Optional[int],
        Option(
            "--route-number",
            "-n",
            help="Individual route to display.",
        ),
    ]

    class Pagination:
        """
        Namespace for pagination options.
        """

        PageNumber = Annotated[
            int,
            Option(
                "--page",
                help="Page number being fetched.",
                show_choices=False,
                show_default=False,
                metavar="INTEGER",
                min=1,
            ),
        ]
        PageSize = Annotated[
            int,
            Option(
                "--page-size",
                help="Number of results per page.",
                show_default=False,
                metavar="INTEGER",
                min=1,
                max=25,
            ),
        ]
        Status = Annotated[
            Optional[Status],
            Option(
                "--status",
                help="Optional filter for matching status.",
            ),
        ]
        EngId = Annotated[
            Optional[str],
            Option(
                "--eng-id",
                help="Optional filter for matching retrosynthesis engine.",
                callback=_eng_id_pagination_callback,
            ),
        ]
        LibId = Annotated[
            Optional[str],
            Option(
                "--lib-id",
                help="Optional filter for matching building block library.",
                callback=_lib_id_pagination_callback,
            ),
        ]

    class Parameters:
        """
        Namespace for retrosynthesis parameter options.
        """

        EngineId = Annotated[
            Optional[str],
            Option(
                "--eng-id",
                help="Retrosynthesis engine ID. Defaults to primary engine.",
                callback=_engine_callback,
            ),
        ]
        LibraryIds = Annotated[
            Optional[List[str]],
            Option(
                "--lib-id",
                help="Building block library IDs. Defaults to all available libraries.",
                callback=_library_callback,
            ),
        ]
        NumRoutes = Annotated[
            int,
            Option(
                "--num-routes",
                help="Maximum number of retrosynthetic routes to generate. Defaults to 20.",
                show_default=False,
                metavar="INTEGER",
                min=1,
                max=50,
            ),
        ]
        ProcessingTime = Annotated[
            int,
            Option(
                "--time-limit",
                help="Maximum processing time in seconds. Defaults to 300.",
                show_default=False,
                metavar="INTEGER",
                min=60,
                max=600,
            ),
        ]
        ReactionLimit = Annotated[
            int,
            Option(
                "--reaction-limit",
                help="Maximum number of times a specific reaction can appear in generated retrosynthetic routes.",
                show_default=False,
                metavar="INTEGER",
                min=1,
                max=25,
            ),
        ]
        BlockLimit = Annotated[
            int,
            Option(
                "--block-limit",
                help="Maximum number of times a building block can appear in a single retrosynthetic route.",
                show_default=False,
                metavar="INTEGER",
                min=1,
                max=25,
            ),
        ]
        JobTags = Annotated[
            List[str],
            Option(
                "--tag",
                help="Up to 3 job tags.",
                callback=_job_tags_callback,
            ),
        ]
        Tags = Annotated[
            List[str],
            Option(
                "--tag",
                help=f"Up to {_JOB_TAG_LIMIT} job tags.",
                callback=_tags_callback,
            ),
        ]



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import sys
from enum import Enum, unique
from typing import Callable, Dict, Optional
from urllib.parse import urljoin

from httpx import Client as HttpxClient, ConnectError, ConnectTimeout, Response
from rich import print


@unique
class _StatusCode(int, Enum):
    """
    Mapping of status code values. Status codes match expected values
    but relate to response codes from the api gateway service from aws.
    """

    BAD_REQUEST = 400
    UNAUTHORIZED = 401
    ACCESS_DENIED = 403
    NOT_FOUND = 404
    PAYLOAD_TOO_LARGE = 413
    UNSUPPORTED_MEDIA_TYPE = 415
    TOO_MANY_REQUESTS = 429
    INTERNAL_SERVER_ERROR = 500
    TIMEOUT = 504


class Client:
    """
    App httpx client wrapper for controlling requests to service apis.
    Method overloaded functions have explicit error handling to exit
    app logic gracefully.

    Args:
        domain (str): Authentication domain url.
        subdomain (str): Application service subdomain url path extenion
            from domain parameter.
        token (str, optional): Authentication access token for target
            domain and application audience.
    """

    _timeout: int = 10

    def __init__(self, *, domain: str, subdomain: str, token: Optional[str] = None):
        self._domain: str = domain
        self._subdomain: str = subdomain
        self._token: Optional[str] = token
        self._client: HttpxClient = self._setup_client()

    def __del__(self) -> None:
        """
        Closing HTTP client when the proxy client is OOM on deleted
        from memory. Reduces handing connections for repeated calls.
        """
        self._client.close()

    def _setup_client(self) -> HttpxClient:
        """
        Setup an httpx client for a base_path url with a global domain
        and specific service domain url path. An access token is added
        as a global header if provided to the client.

        Returns:
            httpx.Client: Initialised httpx client.
        """
        headers: Dict = {}
        if self._token is not None:
            headers["Authorization"] = f"Bearer {self._token}"
        base_url: str = urljoin(self._domain, self._subdomain)
        return HttpxClient(base_url=base_url, headers=headers, timeout=self._timeout)

    @staticmethod
    def _capture_request_errors(method: Callable) -> Callable:
        """
        Capture exceptions and errors that occur when interfacing with
        the client API connection.

        Args:
            method (Callable): HTTP method client function that is being
                wrapped in required error handling.

        Returns:
            Callable: Decorator method with client error handling.
        """

        def request_wrapper(*args, **kwargs) -> Response:
            try:
                skip_errors: bool = kwargs.pop("skip_errors", False)
                response: Response = method(*args, **kwargs)

                # Capture known status code errors from the response
                # when it returns expectedly, only return if none of the
                # known codes are received.

                if skip_errors:
                    return response
                if response.status_code == _StatusCode.BAD_REQUEST:
                    print(
                        "[red]\u2717 (400)",
                        "[red]Request parameters are invalid, contact support for more",
                        "[red]information or see service documentation.",
                    )
                elif response.status_code == _StatusCode.UNAUTHORIZED:
                    print(
                        "[red]\u2717 (401)",
                        "[red]Unauthorized access, try re-authenticating by logging out",
                        "[red]or contact support for more information.",
                    )
                elif response.status_code == _StatusCode.ACCESS_DENIED:
                    if "customer_not_subscribed" in response.content.decode():
                        print(
                            "[red]\u2717",
                            "[red]You are not subscribed to this service, contact the",
                            "[red]Pending AI support team to sign up now.",
                        )
                    else:
                        print(
                            "[red]\u2717 (403)",
                            "[red]Access denied for the service, try logging out or",
                            "[red]contact support for more information.",
                        )
                elif response.status_code == _StatusCode.NOT_FOUND:
                    print(
                        "[red]\u2717 (404)",
                        "[red]Server resource not found or is unavailable, contact",
                        "[red]support for more information.",
                    )
                elif response.status_code == _StatusCode.PAYLOAD_TOO_LARGE:
                    print(
                        "[red]\u2717 (413)",
                        "[red]Payload too large, reduce size of the request or",
                        "[red]contact support for more information.",
                    )
                elif response.status_code == _StatusCode.UNSUPPORTED_MEDIA_TYPE:
                    print(
                        "[red]\u2717 (415)",
                        "[red]Server received an unsupported media type, check your",
                        "[red]request or contact support for more information.",
                    )
                elif response.status_code == _StatusCode.TOO_MANY_REQUESTS:
                    print(
                        "[red]\u2717 (429)",
                        "[red]Server received too many requests, reduce frequency of",
                        "[red]requests or contact support for more information.",
                    )
                elif response.status_code == _StatusCode.INTERNAL_SERVER_ERROR:
                    print(
                        "[red]\u2717 (500)",
                        "[red]Server error encountered, try again shortly or contact",
                        "[red]support for more information.",
                    )
                elif response.status_code == _StatusCode.TIMEOUT:
                    print(
                        "[red]\u2717 (504)",
                        "[red]Server timed out, try again shortly or contact support",
                        "[red]for more information.",
                    )
                else:
                    return response

            # Exception handling for client connection errors during
            # initial setup, or capture other unexpected runtime errors.

            except ConnectError:
                print(
                    "[red]\u2717 (501)",
                    "[red]Connection failed, try again shortly or contact support",
                    "[red]for more information.",
                )
            except ConnectTimeout:
                print(
                    "[red]\u2717 (502)",
                    "[red]Connection timed out, try again shortly or contact support",
                    "[red]for more information.",
                )
            except Exception:
                print(
                    "[red]\u2717 (503)",
                    "[red]Server failed unexpectedly, try again shortly or contact",
                    "[red]support for more information.",
                )

            # Exit application since client failed to respond

            sys.exit(1)

        return request_wrapper

    @_capture_request_errors
    def get(self, *args, **kwargs) -> Response:
        """
        Client method wrapper for making a `GET` request.

        Returns:
            Response: Error-handling HTTP response.
        """
        return self._client.get(*args, **kwargs)

    @_capture_request_errors
    def post(self, *args, **kwargs) -> Response:
        """
        Client method wrapper for making a `POST` request.

        Returns:
            Response: Error-handling HTTP response.
        """
        return self._client.post(*args, **kwargs)

    @_capture_request_errors
    def patch(self, *args, **kwargs) -> Response:
        """
        Client method wrapper for making a `PATCH` request.

        Returns:
            Response: Error-handling HTTP response.
        """
        return self._client.patch(*args, **kwargs)

    @_capture_request_errors
    def put(self, *args, **kwargs) -> Response:
        """
        Client method wrapper for making a `PUT` request.

        Returns:
            Response: Error-handling HTTP response.
        """
        return self._client.put(*args, **kwargs)

    @_capture_request_errors
    def delete(self, *args, **kwargs) -> Response:
        """
        Client method wrapper for making a `DELETE` request.

        Returns:
            Response: Error-handling HTTP response.
        """
        return self._client.delete(*args, **kwargs)

    @_capture_request_errors
    def options(self, *args, **kwargs) -> Response:
        """
        Client method wrapper for making a `OPTIONS` request.

        Returns:
            Response: Error-handling HTTP response.
        """
        return self._client.options(*args, **kwargs)

    @_capture_request_errors
    def head(self, *args, **kwargs) -> Response:
        """
        Client method wrapper for making a `HEAD` request.

        Returns:
            Response: Error-handling HTTP response.
        """
        return self._client.head(*args, **kwargs)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

from pendingai.auth.flows import DeviceAuthorizationFlow, RefreshTokenFlow
from pendingai.auth.models import (
    AccessToken,
    access_token_is_expired,
    get_jwt_token_claims,
)

__all__ = [
    "RefreshTokenFlow",
    "DeviceAuthorizationFlow",
    "AccessToken",
    "access_token_is_expired",
    "get_jwt_token_claims",
]



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

from __future__ import annotations

import time
from collections import defaultdict
from typing import Any, Dict, Generator, List, Optional

from httpx import Response
from pendingai.commands.controller import Controller, capture_controller_errors
from pendingai.commands.retrosynthesis.models import (
    Engine,
    Job,
    JobPage,
    Library,
    NewJob,
)
from rich import print
from rich.console import Console
from rich.progress import track
from rich.status import Status
from typer import Exit

_BATCH_SIZE: int = 100


class Retrosynthesis(Controller):
    """
    Retrosynthesis service subcommand controller.
    """

    @capture_controller_errors
    def get_engines(self) -> List[Engine]:
        """
        Get all available retrosynthesis engine instances for a user.

        Returns:
            List[Engine]: Retrosynthesis engines for the user.
        """
        return [
            Engine.model_validate(x) for x in self.ctx.obj.client.get("/engines").json()
        ]

    @capture_controller_errors
    def get_libraries(self) -> List[Library]:
        """
        Get all available retrosynthesis libraries instances for a user.

        Returns:
            List[Library]: Retrosynthesis libraries for the user.
        """
        return [
            Library.model_validate(x)
            for x in self.ctx.obj.client.get("/libraries").json()
        ]

    @capture_controller_errors
    def get_paginated_jobs(
        self,
        *,
        page: int,
        page_size: int,
        status: Optional[str],
        tags: Optional[List[str]],
        console: Console = Console(),
    ) -> JobPage:
        """
        Retrieve a paginated view of retrosynthesis jobs for a set of
        filter arguments.

        Args:
            page (int): Page number to fetch.
            page_size (int): Page size for the response.
            status (str, optional): Filter value for status.
            tags (List[str], optional): Filter value for tags.
            console (Console, optional): Console for user feedback.

        Returns:
            List[Job]: Retrosynthesis job page.
        """
        params: Dict = {
            "page": page,
            "page_size": page_size,
            "tags": tags,
        }
        if status is not None:
            params["status"] = status

        url: str = "/queries"
        res: Response = self.ctx.obj.client.get(url, params=params, skip_errors=True)

        if res.status_code == 200:
            page_data: JobPage = JobPage.model_validate(res.json())
            page_data.results.sort(key=lambda x: x.created)
            return page_data

        elif res.status_code == 404:
            console.print("[yellow]! No results found with the given filter.")
            raise Exit(0)

        else:
            console.print("[red]\u2717 Failed retrieving jobs. Try again shortly.")
            raise Exit(1)

    @capture_controller_errors
    def submit_molecules(
        self,
        targets: List[str],
        *,
        engine: Optional[str],
        libraries: Optional[List[str]],
        number_of_routes: int,
        processing_time: int,
        reaction_limit: int,
        building_block_limit: int,
        tags: List[str],
        console: Console = Console(),
    ) -> Generator[List[NewJob], Any, None]:
        """
        Submit a collection of SMILES molecules with a set of optional
        parameters to an optional retrosynthesis engine.

        Args:
            targets (List[str]): Collection of query structures.
            engine (str, optional): Retrosynthesis engine.
            libraries (List[str], optional): Building block libraries.
            number_of_routes (int): Number of routes parameter.
            processing_time (int): Processing time parameter.
            reaction_limit (int): Reaction limit parameter.
            building_block_limit (int): Building block limit parameter.
            tags (List[str], optional): Up to 5 job tags.
            console (Console, optional): Console for user feedback.

        Raises:
            BadParameter: Invalid number of targets.
            BadParameter: Invalid number of tags.
            NotImplementedError: Unexpected status code.

        Returns:
            List[NewJob]: Submitted molecule queries and created IDs.
        """

        # Build the shared parameter set for submitting the jobs to
        # the retrosynthesis API. Some parameters need to be dynamically
        # added since null fields are not allowed.
        parameters: dict = {
            "tags": tags,
            "parameters": {
                "number_of_routes": number_of_routes,
                "processing_time": processing_time,
                "reaction_limit": reaction_limit,
                "building_block_limit": building_block_limit,
            },
        }

        if engine is not None:
            parameters["parameters"]["retrosynthesis_engine"] = engine
        else:
            console.print("[green]\u2713[/] Added default retrosynthesis engine.")

        if libraries is not None:
            parameters["parameters"]["building_block_libraries"] = libraries
        else:
            console.print("[green]\u2713[/] Added default building block libraries.")

        # Iterate over the targets in mini-capped batches to not over-
        # whelm the API endpoint. A small 0.5 second wait time is also
        # added to reduce load put into the endpoint and other services.
        batches: int = len(targets) // _BATCH_SIZE + 1
        for i in range(0, len(targets), _BATCH_SIZE):
            result: List[NewJob]
            request: Dict = parameters | {"query": targets[i : i + _BATCH_SIZE]}
            batch_size: int = i // _BATCH_SIZE + 1
            with Status(f"Submitting batch {batch_size} of {batches}", console=console):
                response: Response = self.ctx.obj.client.post("/queries", json=request)
                if response.status_code == 200:
                    result = [NewJob.model_validate(x) for x in response.json()]
                elif response.status_code == 202:
                    target_id: str = response.headers.get("Location").split("/")[-2]
                    result = [NewJob(id=target_id, smiles=targets[i])]
                else:
                    print("[red]\u2717[/] Failed to submit jobs, try again shortly.")
                    raise Exit(1)
            console.print(f"[green]\u2713[/] Submitted batch {batch_size} of {batches}")
            yield result
            time.sleep(0.5)

    @capture_controller_errors
    def delete_jobs(
        self,
        ids: List[str],
        *,
        console: Console = Console(),
    ) -> Dict[str, List[str]]:
        """
        Delete a list of input list and file job IDs.

        Args:
            ids (List[str]): Input list with job IDs.
            console (Console, optional): Console for user feedback.

        Returns:
            Tuple[List[str], List[str], List[str]]: Lists of successful,
                missing and failed job ID deletion operations.
        """
        results: Dict[str, List[str]] = defaultdict(list)
        for i in track(ids, "Deleting jobs", transient=True, console=console):
            res: Response = self.ctx.obj.client.delete(f"/queries/{i}", skip_errors=True)
            if res.status_code == 204:
                results["success"].append(i)
            elif res.status_code == 404:
                results["missing"].append(i)
            else:
                results["failure"].append(i)
        return results

    @capture_controller_errors
    def get_job_status(
        self,
        ids: List[str],
        *,
        console: Console = Console(),
    ) -> List[Dict[str, str]]:
        """
        Get the status of a list of input list and file job IDs.

        Args:
            ids (List[str]): Input list with job IDs.
            console (Console, optional): Console for user feedback.

        Returns:
            List[Dict[str, str]]: Status matching the job IDs.
        """
        results: List[Dict] = []
        for i in track(ids, "Check job statuses", transient=True, console=console):
            endpoint: str = f"/queries/{i}/status"
            res: Response = self.ctx.obj.client.get(endpoint, skip_errors=True)
            if res.status_code == 200:
                results.append({"id": i, "status": res.json()["status"]})
            elif res.status_code == 303:
                results.append({"id": i, "status": "completed"})
            elif res.status_code == 404:
                results.append({"id": i, "status": "missing"})
            else:
                results.append({"id": i, "status": "failed"})
        return sorted(results, key=lambda x: x["status"])

    @capture_controller_errors
    def get_jobs_by_id(
        self,
        ids: List[str],
        *,
        console: Console = Console(),
    ) -> Generator[Job | Dict, Any, None]:
        """
        Retrieve jobs by ID.

        Args:
            ids (List[str]): Input list with job IDs.
            console (Console, optional): Console for user feedback.

        Raises:
            BadParameter: Job ID does not exist.

        Returns:
            List[Job]: Job results.
        """
        for i in track(ids, "Retrieving results", transient=True, console=console):
            res: Response = self.ctx.obj.client.get(f"/queries/{i}", skip_errors=True)
            if res.status_code == 200:
                yield Job.model_validate(res.json() | {"id": i})
            else:
                yield {"id": i, "result": None}

    @capture_controller_errors
    def get_job_tags(self) -> list[str]:
        """
        Retrieve all jobs tags.

        Returns:
            list[str]: Sorted list of tags.
        """
        response: Response = self.ctx.obj.client.get("/query_tags")
        if response.status_code == 200:
            return sorted(response.json())
        return []

    @capture_controller_errors
    def get_job_tag_count(self, tag_id: str) -> int:
        """
        Retrieve count of jobs for a tag.

        Args:
            tag_id (str): Tag to count.

        Returns:
            int: Count of jobs for a tag.
        """
        response: Response = self.ctx.obj.client.get(f"/query_tags/{tag_id}/count")
        if response.status_code == 200:
            return response.json().get("count", 0)
        return 0

    @capture_controller_errors
    def delete_job_tag(self, tag_id: str) -> int:
        """
        Delete all jobs for a tag.

        Args:
            tag_id (str): Tag to delete.

        Returns:
            int: Count of deleted jobs for a tag.
        """
        response: Response = self.ctx.obj.client.delete(f"/query_tags/{tag_id}")
        if response.status_code == 200:
            return response.json().get("count", 0)
        return 0



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import enum
import logging.config
import typing

import httpx
from rich.logging import RichHandler

__appname__: str = "pendingai"
__version__: str = "0.1.4"

logging.basicConfig(
    level=logging.WARNING,
    format="%(message)s",
    datefmt="[%X]",
    handlers=[RichHandler(tracebacks_suppress=[httpx])],
)
logging.getLogger("httpcore.connection").setLevel(logging.WARNING)
logging.getLogger("httpcore.http11").setLevel(logging.WARNING)
logging.getLogger("httpx").setLevel(logging.WARNING)


@enum.unique
class Environment(str, enum.Enum):
    """
    Deployment environment used for building client connection strings,
    authentication flows for the device or refresh tokens, and controls
    cached state data between different environments
    """

    DEVELOPMENT = "dev"
    STAGING = "stage"
    PRODUCTION = "prod"


@enum.unique
class Command(str, enum.Enum):
    """
    Runtime command used for naming commands and linking the client
    service domain prefix.
    """

    RETRO = "retro"


__all__: typing.List[str] = ["__appname__", "__version__", "Environment", "Command"]



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import logging
import webbrowser
from typing import Optional

import typer
from rich import print
from typer import Exit, Option, Typer, echo
from typing_extensions import Annotated

import pendingai.commands.retrosynthesis.command as retrosynthesis
from pendingai import Environment, __appname__, __version__
from pendingai.commands.auth.command import command as auth_command
from pendingai.context import Context, Namespace

namespace: Namespace = Namespace.load()

logger: logging.Logger = logging.getLogger("root")


app: Typer = Typer(name=__appname__, no_args_is_help=True, add_completion=False)
app.pretty_exceptions_show_locals = False
app.rich_markup_mode = None

app.add_typer(retrosynthesis.app, name=namespace.commands.retrosynthesis.subcommand)
app.add_typer(auth_command)


def _version_callback(value: bool) -> None:
    """Display app version and exit."""
    if value:
        echo(f"{__appname__}/{__version__}")
        raise Exit(0)


@app.callback()
def _callback(
    ctx: typer.Context,
    version: Annotated[
        bool,
        Option(
            "--version",
            is_flag=True,
            is_eager=True,
            callback=_version_callback,
            help="Show the application version and exit.",
        ),
    ] = False,
    verbosity: Annotated[
        int,
        Option(
            "-v",
            "--verbose",
            hidden=True,
            count=True,
            help="Verbose flag for stdout log-level.",
        ),
    ] = 0,
    environment: Annotated[
        Environment,
        Option(
            "--env",
            "-e",
            hidden=True,
            show_default=False,
            envvar="PENDINGAI_ENVIRONMENT",
            help="Selectable runtime deployment server.",
        ),
    ] = Environment.PRODUCTION,
    api_key: Annotated[
        Optional[str],
        Option(
            "--api-key",
            hidden=True,
            envvar="PENDINGAI_API_KEY",
            help="Session token alternative API key.",
        ),
    ] = None,
) -> None:
    """
    Pending AI Command-Line Interface.

    Cheminformatics services offered by this CLI are accessible through an
    API integration with Pending AI servers. An authenticated session is
    required for use; see <pendingai auth>. More information is available
    with <pendingai docs>.
    """
    logger.setLevel(40 - min(verbosity, 4) * 10)
    ctx.obj = Context(
        app_name=__appname__,
        environment=environment,
        api_key=api_key,
        subcommand=ctx.invoked_subcommand,
    )


@app.command("docs")
def docs(ctx: typer.Context) -> None:
    """
    Open the Pending AI documentation page in a web browser.
    """
    prompt: str = "[bold]Press Enter[/] to open %s in your browser. "
    print(prompt % namespace.docs_redirect_url, end="")
    input()
    webbrowser.open_new_tab(namespace.docs_redirect_url)


@app.command("support")
def support(ctx: typer.Context) -> None:
    """
    Open the Pending AI support page in a web browser.
    """
    print("[yellow]! Lodge a ticket via email: [blue underline]support@pending.ai")
    print("[yellow]- Refer to [b]<pendingai docs>[/] for expected errors.")
    print("[yellow]- Provide information such as the command and any output provided.")
    print("[yellow]- Please use the same email address registered with your account.")


if __name__ == "__main__":
    app()



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import sys
import time
import webbrowser
from abc import ABC
from typing import Any, Dict, Optional

import httpx
from pydantic import ValidationError
from rich import print
from rich.status import Status

from pendingai.auth.models import AccessToken


class _AuthenticationFlow(ABC):
    """
    Abstract authentication flow controller class. Enforces flows to use
    a single authentication domain, client ID and audience for handling
    flow API requests. Flows provide utility methods that may need to be
    called in sequence / separately with additional output.

    Args:
        domain (str): Authentication client domain.
        client_id (str): Authentication client ID within the domain.
        audience (str): Authentication client audience for the domain.
    """

    def __init__(self, domain: str, client_id: str, audience: str):
        self._domain: str = domain
        self._client_id: str = client_id
        self._audience: str = audience


class RefreshTokenFlow(_AuthenticationFlow):
    """
    Refresh token flow from Auth0 is used to re-authenticate an already
    expired access token. The `offline_access` grant is used by the API
    application and will eventually invalidate refresh tokens that have
    exceeded a usage period.

    See more: https://auth0.com/docs/secure/tokens/refresh-tokens
    """

    def refresh(self, refresh_token: str) -> Optional[AccessToken]:
        """
        Attempts to rotate a refresh token. Any error encountered yields
        a null response indicating a failed refresh flow.

        Args:
            refresh_token (str): Previously generated refresh token for
                the matching authentication domain.

        Returns:
            Optional[AccessToken]: Successfully rotated access token or
                null from a failed refresh flow.
        """
        with Status("Refreshing device access"):
            response: httpx.Response = httpx.post(
                url=f"https://{self._domain}/oauth/token",
                data={
                    "client_id": self._client_id,
                    "audience": self._audience,
                    "grant_type": "refresh_token",
                    "refresh_token": refresh_token,
                },
            )

        if response.status_code == 200:
            return AccessToken.model_validate(response.json())
        return None


class DeviceAuthorizationFlow(_AuthenticationFlow):
    """
    Device authorization flow from Auth0 is used to authenticate a user
    from the command-line interface. Both `openid` and `offline_access` auth
    scopes are used for detailed access token data. The process requires
    explicit user input so additional stdout is given when called.

    See more: https://auth0.com/docs/get-started/authentication-and-authorization-flow/device-authorization-flow
    """

    _max_retries: int = 25

    def _generate_device_code(self) -> Dict:
        """
        Utility method for requesting and parsing the authentication
        response when generating a device authorization code. Additional
        output is given to reflect the runtime status.

        Returns:
            Dict: Response body data.
        """
        with Status("Requesting device authorization code"):
            response: httpx.Response = httpx.post(
                url=f"https://{self._domain}/oauth/device/code",
                data={
                    "client_id": self._client_id,
                    "audience": self._audience,
                    "scope": "openid profile email offline_access",
                },
            )

        if response.status_code == 200:
            print("[green]\u2713[/] Device authorization code received.")
            return response.json()
        if response.status_code >= 400:
            print("[red]\u2717 Authentication service failed unexpectedly.")
        else:
            print("[red]\u2717 Failed requesting device code, try again.")
        sys.exit(1)

    def _generate_access_token(self, device_code: str) -> Optional[AccessToken]:
        """
        Utility method for requesting the access token on authentication
        server periodic checks based on an interval device code time. An
        access code is received successfully once user is registered.

        Args:
            device_code (str): Pre-generated device authorization code.

        Returns:
            Optional[AccessToken]: Parsed access token data model when
                successfully authenticated, otherwise null.
        """
        response: httpx.Response = httpx.post(
            url=f"https://{self._domain}/oauth/token",
            data={
                "client_id": self._client_id,
                "audience": self._audience,
                "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
                "device_code": device_code,
            },
        )

        if response.status_code == 200:
            try:
                return AccessToken.model_validate(response.json())
            except ValidationError:
                print("[red]\u2717 Unexpected session token received, try again.")

        elif response.status_code == 403:
            if response.json()["error"] == "authorization_pending":
                return None
            if response.json()["error"] == "expired_token":
                print("[red]\u2717 Session was closed by the authentication server.")

        elif response.status_code == 429:
            print("[yellow]! Too many requests received by authentication server.")

        else:
            print("[red]\u2717 Failed requesting session token, try again.")

        sys.exit(1)

    def authorize(self) -> AccessToken:
        """
        Executes the interactive device authorization flow. Generate a
        device code session with Auth0, open a webbrowser instance and
        ping repeatedly until the device has been authenticated.
        """

        # Generate device code for authorization redirect url
        device_code_data: Dict[str, Any] = self._generate_device_code()
        redirect: str = device_code_data["verification_uri_complete"]
        user_code: str = device_code_data["user_code"]
        device_code: str = device_code_data["device_code"]

        print(f"- Navigate to the url on your device: {redirect}")
        print(f"- Enter the following code: [b]{user_code}[/]")
        time.sleep(2)
        webbrowser.open_new_tab(redirect)

        # Generate access token if user has authenticated logged in
        with Status("Waiting for device authentication"):
            for _ in range(self._max_retries):
                token: Optional[AccessToken] = self._generate_access_token(device_code)
                if token is not None:
                    return token
                time.sleep(device_code_data["interval"])

        # Capture authentication timeout on max retries
        print("[red]\u2717 Authentication timed out, try again.")
        sys.exit(1)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import logging
from pathlib import Path
from typing import Any, Dict, Optional

import typer
from pydantic import BaseModel, ValidationError
from yaml import safe_load

from pendingai import Environment, __appname__
from pendingai.auth.models import AccessToken

logger: logging.Logger = logging.getLogger("root")


class Cache(BaseModel):
    """
    Cache file data model.
    """

    access_token: Optional[AccessToken] = None

    @classmethod
    def load(cls, environment: Environment) -> "Cache":
        """
        Load cache runtime based on a given environment.

        Args:
            environment (Environment): Runtime environment.

        Returns:
            Cache: Loaded class instance.
        """
        try:
            contents: Dict[str, Any] = {}
            file: str = f".{environment.value}.cache"
            path: Path = Path(typer.get_app_dir(__appname__, force_posix=True)) / file
            if not path.parent.exists():
                logger.info(f"Initialising cache directory: {path.parent}")
                path.parent.mkdir()
            if not path.parent.is_dir():
                logger.warning(f"Invalid cache directory: {path.parent}")
                logger.warning("Unable to initialise cache instance, using default.")
                return cls.model_validate(contents)
            if path.exists():
                if path.is_file():
                    return cls.model_validate(safe_load(path.open()))
                logger.warning(f"Invalid cache filepath: {path}")
                logger.warning("Unable to initialise cache instance, using default.")
            return cls.model_validate(contents)

        except ValidationError:
            logger.warning("App cache contains malformed data, clearing invalid data.")
            logger.warning("Unable to initialise cache instance, using default.")
            return cls.model_validate({})

    def save(self, environment: Environment) -> None:
        """
        Save cache contents to file.

        Args:
            environment (Environment): Saved cache environment.
        """
        file: str = f".{environment.value}.cache"
        path: Path = Path(typer.get_app_dir(__appname__, force_posix=True)) / file
        with path.open("w") as fp:
            fp.write(self.model_dump_json())

    def delete(self, environment: Environment) -> None:
        """
        Delete cache content file.

        Args:
            environment (Environment): Cache environment being deleted.
        """
        file: str = f".{environment.value}.cache"
        path: Path = Path(typer.get_app_dir(__appname__, force_posix=True)) / file
        path.unlink(missing_ok=True)
